import unittest


class StateMachineTests(unittest.TestCase):

    def _getTargetClass(self):
        from repoze.workflow.statemachine import StateMachine
        return StateMachine

    def _makeOne(self, attr='state', **kw):
        klass = self._getTargetClass()
        return klass(attr, **kw)

    def test_add(self):
        sm = self._makeOne()
        sm.add('start', 'add', 'adding', None)
        sm.add('start', 'get', 'getting', None)
        self.assertEqual(sm.states[('start', 'add')], ('adding', None, {}))
        self.assertEqual(sm.states[('start', 'get')], ('getting', None, {}))

    def test_add_with_kw(self):
        sm = self._makeOne()
        sm.add('start', 'add', 'adding', None, a=1)
        sm.add('start', 'get', 'getting', None, b=2)
        self.assertEqual(sm.states[('start', 'add')], ('adding', None, {'a':1}))
        self.assertEqual(sm.states[('start', 'get')],('getting', None, {'b':2}))

    def test_add_in_separate_instance_does_not_pollute(self):
        sm1 = self._makeOne()
        sm2 = self._makeOne()
        sm1.add('start', 'add', 'adding', None)
        sm1.add('start', 'get', 'getting', None)
        self.assertEqual(len(sm2.states), 0)

    def test_transitions(self):
        sm = self._makeOne(initial_state='pending')
        sm.add('pending', 'publish', 'published', None)
        sm.add('pending', 'reject', 'private', None)
        sm.add('published', 'retract', 'pending', None)
        sm.add('private', 'submit', 'pending', None)
        sm.add('pending', None, 'published', None)
        ob = ReviewedObject()
        self.assertEqual(sorted(sm.transitions(ob)), [None,'publish', 'reject'])
        self.assertEqual(sorted(sm.transitions(ob, from_state='private')),
                         ['submit'])
        ob.state = 'published'
        self.assertEqual(sorted(sm.transitions(ob)), ['retract'])

    def test_transition_info(self):
        sm = self._makeOne(initial_state='pending')
        sm.add('pending', 'publish', 'published', None, a=1)
        sm.add('pending', 'reject', 'private', None, b=2)
        sm.add('published', 'retract', 'pending', None, c=3)
        sm.add('private', 'submit', 'pending', None, d=4)
        sm.add('pending', None, 'published', None, e=5)
        ob = ReviewedObject()
        info = sorted(sm.transition_info(ob))
        self.assertEqual(len(info), 3)
        self.assertEqual(
            info[0],
            {'a': 1, 'from_state': 'pending',
             'to_state': 'published', 'transition_id': 'publish'})
        self.assertEqual(
            info[1],
            {'b':2, 'from_state': 'pending',
             'to_state': 'private', 'transition_id': 'reject'})
        self.assertEqual(
            info[2],
            {'e':5, 'from_state': 'pending',
             'to_state': 'published', 'transition_id': None})
        ob.state = 'published'
        info = sorted(sm.transition_info(ob))
        self.assertEqual(len(info), 1)
        self.assertEqual(
            info[0],
             {'c': 3, 'from_state': 'published',
              'to_state': 'pending', 'transition_id': 'retract'})

    def test_execute_use_add(self):
        sm = self._makeOne(initial_state='pending')
        args = []
        def dummy(state, newstate, transition_id, context, **kw):
            args.append((state, newstate, transition_id, context, kw))
        sm.add('pending', 'publish', 'published', dummy, a=1)
        sm.add('pending', 'reject', 'private', dummy, b=2)
        sm.add('published', 'retract', 'pending', dummy, c=3)
        sm.add('private', 'submit', 'pending', dummy, d=4)
        sm.add('pending', None, 'published', dummy, e=5)
        ob = ReviewedObject()
        sm.execute(ob, 'publish')
        self.assertEqual(ob.state, 'published')
        sm.execute(ob, 'retract')
        self.assertEqual(ob.state, 'pending')
        sm.execute(ob, 'reject')
        self.assertEqual(ob.state, 'private')
        sm.execute(ob, 'submit')
        self.assertEqual(ob.state, 'pending')
        # catch-all
        sm.execute(ob, None)
        self.assertEqual(ob.state, 'published')
        self.assertEqual(len(args), 5)
        self.assertEqual(args[0],
                         ('pending', 'published', 'publish', ob, {'a':1}))
        self.assertEqual(args[1],
                         ('published', 'pending', 'retract', ob, {'c':3}))
        self.assertEqual(args[2],
                         ('pending', 'private', 'reject', ob, {'b':2}))
        self.assertEqual(args[3],
                         ('private', 'pending', 'submit', ob, {'d':4}))
        self.assertEqual(args[4],
                         ('pending', 'published', None, ob, {'e':5}))
        from repoze.workflow.statemachine import StateMachineError
        self.assertRaises(StateMachineError, sm.execute, ob, 'nosuch')

    def test_execute_use_constructor(self):
        args = []
        def dummy(state, newstate, transition_id, context, **kw):
            args.append((state, newstate, transition_id, context, kw))
        states = {('pending', 'publish'): ('published', dummy, {}),
                  ('pending', 'reject'): ('private', dummy, {}),
                  ('published', 'retract'): ('pending', dummy, {}),
                  ('private', 'submit'): ('pending', dummy, {}),
                  ('pending', None): ('published', dummy, {}),}
        sm = self._makeOne(states=states, initial_state='pending')
        ob = ReviewedObject()
        sm.execute(ob, 'publish')
        self.assertEqual(ob.state, 'published')
        sm.execute(ob, 'retract')
        self.assertEqual(ob.state, 'pending')
        sm.execute(ob, 'reject')
        self.assertEqual(ob.state, 'private')
        sm.execute(ob, 'submit')
        self.assertEqual(ob.state, 'pending')
        # catch-all
        sm.execute(ob, None)
        self.assertEqual(ob.state, 'published')
        self.assertEqual(len(args), 5)
        self.assertEqual(args[0], ('pending', 'published', 'publish', ob, {}))
        self.assertEqual(args[1], ('published', 'pending', 'retract', ob, {}))
        self.assertEqual(args[2], ('pending', 'private', 'reject', ob, {}))
        self.assertEqual(args[3], ('private', 'pending', 'submit', ob, {}))
        self.assertEqual(args[4], ('pending', 'published', None, ob, {}))
        from repoze.workflow.statemachine import StateMachineError
        self.assertRaises(StateMachineError, sm.execute, ob, 'nosuch')

    def test_execute_None_match(self):
        sm = self._makeOne(initial_state='pending')
        args = []
        def dummy(state, newstate, transition_id, context, **kw):
            args.append((state, newstate, transition_id, context, kw))
        sm.add('pending', None, 'published', dummy)
        ob = ReviewedObject()
        sm.execute(ob, 'publish')
        self.assertEqual(ob.state, 'published')

    def test_fail_before_transition(self):
        from repoze.workflow.statemachine import StateMachine
        from repoze.workflow.statemachine import StateMachineError
        
        class FailBeforeTransition(StateMachine):
            def before_transition(self, a, b, c, d):
                raise StateMachineError

        def dummy(state, newstate, transition_id, context):
            pass

        states = {('from', 'do_it'):('to', dummy, {})}
        sm = FailBeforeTransition('state', states, initial_state='from')
        ob = ReviewedObject()
        self.assertRaises(StateMachineError, sm.execute, ob, 'do_it')
        self.assertEqual(hasattr(ob, 'state'), False)
        self.assertEqual(sm.state_of(ob), 'from')

class ReviewedObject:
    pass



from zope.interface import Interface

class IStateMachine(Interface):
    def add(state, transition_id, newstate, transition_fn, **kw):
        """Add a transition to the FSM."""

    def execute(context, transition_id):
        """Perform a transition and execute an action."""

    def state_of(context):
        """ Return the current state of the given object """

    def transitions(context, from_state=None):
        """ Return the available transition ids for the given object
        (from_state defaults to the object's current state)"""

    def transition_info(context, from_state=None):
        """ Return sequence of dictionaries representing the
        transition information for context (from_state defaults to the
        object's current state).  Each dictionary has the keys
        ``transition_id``, ``from_state``, ``to_state`` as well as any
        keywords passed in to the ``add`` method for this transition."""
    
    def before_transition(state, newstate, transition_id, context, **kw):
        """
        Hook method to be overridden by subclasses (or injected
        directly onto an instance) to allow for before transition
        actions (such as firing an event).

        Raise an exception here to abort the transition.
        """

    def after_transition(state, newstate, transition_id, context, **kw):
        """
        Hook method to be overridden by subclasses (or injected
        directly onto an instance) to allow for after transition
        actions (such as firing an event).
        """
        



""" Finite state machine, useful for workflow-like features, based on
Skip Montanaro's FSM from
http://wiki.python.org/moin/FiniteStateMachine (ancient but simple #
and useful!)"""
from repoze.workflow.interfaces import IStateMachine
from zope.interface import implements

_marker = ()

class StateMachineError(Exception):
    """ Invalid input to finite state machine"""

class StateMachine(object):
    """ Finite state machine featuring transition actions.

    The class stores a dictionary of (from_state, transition_id) keys, and
    (to_state, transition_fn) values.

    When a (state, transition_id) search is performed:
    * an exact match is checked first,
    * (state, None) is checked next.

    The transition function must be of the following form:
    * function(current_state, new_state, transition_id, context, **kw)

    It is recommended that all transition functions be module level
    callables to facilitate issues related to StateMachine
    persistence.
    """
    implements(IStateMachine)
    
    def __init__(self, state_attr, states=None, initial_state=None):
        """
        o state_attr - attribute name where a given object's current
                       state will be stored (object is responsible for
                       persisting)
                       
        o states - state dictionary

        o initial_state - initial state for any object using this
                          state machine
        """
        if states is None:
            states = {}
        self.states = states
        self.state_attr = state_attr
        self.initial_state = initial_state

    def add(self, state, transition_id, newstate, transition_fn, **kw):
        self.states[(state, transition_id)] = (newstate, transition_fn, kw)

    def execute(self, context, transition_id):
        state = getattr(context, self.state_attr, _marker) 
        if state is _marker:
            state = self.initial_state
        si = (state, transition_id)
        sn = (state, None)
        newstate = None
        # exact state match?
        if si in self.states:
            newstate, transition_fn, kw = self.states[si]
        # no exact match, how about a None (catch-all) match?
        elif sn in self.states:
            newstate, transition_fn, kw = self.states[sn]
        if newstate is None:
            raise StateMachineError(
                'No transition from %r using transition %r'
                    % (state, transition_id))
        self.before_transition(state, newstate, transition_id, context, **kw)
        transition_fn(state, newstate, transition_id, context, **kw)
        self.after_transition(state, newstate, transition_id, context, **kw)
        setattr(context, self.state_attr, newstate)

    def state_of(self, context):
        state = getattr(context, self.state_attr, self.initial_state)
        return state

    def transitions(self, context, from_state=None):
        if from_state is None:
            from_state = self.state_of(context)
        transitions = [t_id for (state, t_id) in self.states.keys()
                       if state == from_state]
        return transitions

    def transition_info(self, context, from_state=None):
        if from_state is None:
            from_state = self.state_of(context)
        L = []
        for (state,t_id), (newstate,transition_fn,kw) in self.states.items():
            if state == from_state:
                newkw = {}
                newkw.update(kw)
                newkw['transition_id'] = t_id
                newkw['from_state'] = state
                newkw['to_state'] = newstate
                L.append(newkw)
        return L

    def before_transition(self, state, newstate, transition_id, context, **kw):
        pass

    def after_transition(self, state, newstate, transition_id, context, **kw):
        pass



##############################################################################
#
# Copyright (c) 2008-2009 Agendaless Consulting and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the BSD-like license at
# http://www.repoze.org/LICENSE.txt.  A copy of the license should accompany
# this distribution.  THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL
# EXPRESS OR IMPLIED WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND
# FITNESS FOR A PARTICULAR PURPOSE
#
##############################################################################

__version__ = '0.4'

import os

from ez_setup import use_setuptools
use_setuptools()

from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README.txt')).read()
CHANGES = open(os.path.join(here, 'CHANGES.txt')).read()

requires = [
    'setuptools',
    'zope.interface>=3.4'
    ]

setup(name='repoze.workflow',
      version=__version__,
      description='Simple state machine useful for workflow-like apps',
      long_description=README + '\n\n' +  CHANGES,
      classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Programming Language :: Python",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "Topic :: Internet :: WWW/HTTP :: WSGI",
        "Topic :: Internet :: WWW/HTTP :: WSGI :: Middleware",
        ],
      keywords='web wsgi zope',
      author="Agendaless Consulting",
      author_email="repoze-dev@lists.repoze.org",
      url="http://www.repoze.org",
      license="BSD-derived (http://www.repoze.org/LICENSE.txt)",
      packages=find_packages(),
      include_package_data=True,
      namespace_packages=['repoze'],
      zip_safe=False,
      tests_require = requires + ['Sphinx', 'repoze.sphinx.autointerface'],
      install_requires=requires,
      test_suite="repoze.workflow",
      entry_points = """\
      """
      )

